use std::fmt::{Display, Formatter, Result};
use actix_web::{ResponseError, error::BlockingError, HttpResponse, http::{header::ContentType, StatusCode}};
use serde::{Serialize, Deserialize};
use validator::ValidationErrors;

use crate::schema::error::DatabaseError;

#[derive(Serialize, Deserialize, Debug)]
pub struct ApiError {
    #[serde(skip_serializing)]
    pub status: u16,
    pub message: String
}

impl Display for ApiError {
    // This trait requires `fmt` with this exact signature.
    fn fmt(&self, f: &mut Formatter) -> Result {
        write!(f, "{}", self.message)
    }
}

impl ResponseError for ApiError {
    fn error_response(&self) -> HttpResponse {
        HttpResponse::build(StatusCode::from_u16(self.status).unwrap())
        .content_type(ContentType::json())
        .body(serde_json::to_string(&self).unwrap())
    }    
}

impl From<DatabaseError> for ApiError {
    fn from(value: DatabaseError) -> Self {
        ApiError {
            //Todo, split this in unrecoverable and schema error
            status: 500,
            message: value.0
        }
    }
}

impl From<BlockingError> for ApiError {
    fn from(value: BlockingError) -> Self {
        ApiError {
            status: 500,
            message: value.to_string()
        }
    }
}

impl From<serde_json::Error> for ApiError {
    fn from(value: serde_json::Error) -> Self {
        ApiError {
            status: 500,
            message: value.to_string()
        }
    }
}

impl From<jsonwebtoken::errors::Error> for ApiError {
    fn from(value: jsonwebtoken::errors::Error) -> Self {
        ApiError {
            status: 500,
            message: value.to_string() 
        }
    }
}

impl From<ValidationErrors> for ApiError {
    fn from(value: ValidationErrors) -> Self {
        ApiError {
            status: 422,
            message: value.to_string()
        }
    }
}

impl From<chrono::ParseError> for ApiError {
    fn from(value: chrono::ParseError) -> Self {
        ApiError {
            status: 422,
            message: value.to_string()
        }
    }
}
